/*
 * Copyright (c) 2003, 2004 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.schemas.dtd;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.InputStreamReader;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;

/**
 * This class implements a DTD catalog for XML DTDs. The DTDs are loaded from
 * the same package directory this class resides in. This allows convenient jar
 * packaging without having to configure paths to the local file system.
 * 
 * <p>
 * The DTDs are identified by public ID. The mappings from public IDs to file
 * names are stored in a file called public-identifiers.cat. This resolver never
 * attempts to open network connections in order to fetch an entity based on the
 * system ID. If an entity doesn't resolve to the local catalog, as zero-length
 * byte stream is returned. This Mozilla-style hack has implications on
 * well-formedness checks. When the XML processor is given a zero length "DTD",
 * it believes it has seen the DTD and treats references to undeclared entities
 * as a fatal errors. The intent is to allow smooth parsing of documents that
 * don't rely on the DTD but still contain a doctype declaration out of habit.
 * 
 * <p>
 * The entity resolver is implemented as a thread-safe singleton.
 * 
 * @author hsivonen
 */
public class DTDCatalog implements EntityResolver {

    /** The mappings from the public IDs to filenames */
    private Map publicIds;

    /** The holder for the singleton instance. */
    private static DTDCatalog instance;

    /**
     * Returns the <code>DTDCatalog</code> instance.
     * 
     * @return the <code>DTDCatalog</code> instance
     */
    public static DTDCatalog getInstance() {
        return instance;
    }

    static {
        instance = new DTDCatalog();
    }

    /** Creates a new instance of DTDCatalog */
    private DTDCatalog() {
        try {
            this.publicIds = new TreeMap();
            Pattern catalogRowPat = Pattern
                    .compile("^\\s*PUBLIC\\s*\"([^\"]*)\"\\s*\"([^\"]*)\"");
            BufferedReader reader = new BufferedReader(new InputStreamReader(
                    this.getClass().getResourceAsStream(
                            "public-identifiers.cat"), "US-ASCII"));
            String line;
            while ((line = reader.readLine()) != null) {
                Matcher m = catalogRowPat.matcher(line);
                if (m.matches()) {
                    this.publicIds.put(m.group(1), m.group(2));
                }
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Resolves an entity based on the public ID.
     * 
     * @param publicId
     *            the public ID of the external entity
     * @param systemId
     *            ignored
     * 
     * @return an <code>InputSource</code> encapsulating the entity if the
     *         entity was found in the local catalog or an
     *         <code>InputSource</code> encapsulating a zero-length byte
     *         stream if the entity was not found in the local catalog
     */
    public InputSource resolveEntity(String publicId, String systemId)
            throws org.xml.sax.SAXException, java.io.IOException {
        String fileName;
        synchronized (this.publicIds) {
            fileName = (String) this.publicIds.get(publicId);
        }
        InputSource is;
        if (fileName != null) {
            is = new InputSource(this.getClass().getResourceAsStream(fileName));
        } else {
            is = new InputSource(new ByteArrayInputStream(new byte[0]));
        }
        is.setPublicId(publicId);
        is.setSystemId(systemId);
        return is;
    }
}