/*
 * Copyright (c) 2003, 2004 Henri Sivonen
 *
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

package fi.iki.hsivonen.template;
import java.io.IOException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Stack;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import fi.iki.hsivonen.xml.DOMUtils;
import fi.iki.hsivonen.xml.HtmlSerializer;
import gnu.xml.pipeline.NSFilter;
import gnu.xml.pipeline.TextConsumer;
import gnu.xml.util.DomParser;

/**
 * An engine for combining an XML template and XHTML documents.
 *
 * <p>Known limitation: <code>dir</code> and <code>xml:lang</code> don't get
 * any smart treatment. This may be problematic if the language and dominant
 * writing direction of the template and the content don't match.
 *
 * @version $Id: TemplateEngine.java,v 1.64 2004/08/30 17:12:28 hsivonen Exp $
 * @author  hsivonen
 */
public class TemplateEngine {

    /** The namespace URL for XHTML */
    final static String XHTML_NS = "http://www.w3.org/1999/xhtml";
    /** The namespace URL for the template namespace */
    final static String TEMPLATE_NS = "http://iki.fi/hsivonen/template";

    private DataSource dataSource;

    private Stack documentStack;

    private Document template;

    private Date lastModified;

    private Response response;

    private URL context;

    protected boolean avoidFullContent;

    protected Dependencies dependencies;

    public TemplateEngine(DataSource dataSource)
        throws DataSourceException, IOException {
        this.dataSource = dataSource;
        this.avoidFullContent = dataSource.shouldAvoidFullContent();
        this.dependencies = dataSource.newDependencies();
    }

    /**
     *
     * @throws ResourceNotFoundException if the main content was not found
     * @throws TemplateException if the <code>TemplateEngine</code> discovered 
     *                           erroneous data
     * @throws IOException if an I/O error occurred
     * @throws DataSourceException if the <code>DataSource</code> encountered
     *                             other difficulty
     */
    public synchronized void respond(URL url, Response response)
        throws TemplateException, IOException, DataSourceException {
        this.lastModified = null;
        this.response = response;
        this.context = url;
        Template temp =
            this.dataSource.getTemplateForURL(url, this.dependencies);
        if (temp == null) {
            // pass through
            this.passThrough();
        } else {
            // we've got a template
            this.template = temp.getDocument();
            this.lastModified = temp.getDateModified();
            if (this.lastModified == null) {
                this.lastModified = new Date(System.currentTimeMillis());
            }
            this.documentStack = new Stack();
            this.processTemplate();
        }
    }

    /** Getter for property dataSource.
     * @return Value of property dataSource.
     *
     */
    protected DataSource getDataSource() {
        return dataSource;
    }

    /** Getter for property currentItem.
     * @return Value of property currentItem.
     *
     */
    protected final Item getCurrentItem()
        throws TemplateException, DataSourceException, IOException {
        if (this.documentStack.isEmpty()) {
            Item result =
                this.getDataSource().select(this.context, this.dependencies);
            if (result == null) {
                throw new ResourceNotFoundException("No document bound to URL");
            }
            this.pushCurrentItem(result);
            this.maybeSetLastModified(result.getDateModified());
        }
        return (Item)this.documentStack.peek();
    }

    protected final void pushCurrentItem(Item item) {
        this.documentStack.push(item);
    }

    protected final void popCurrentItem() {
        this.documentStack.pop();
    }

    protected final void maybeSetLastModified(Date modified) {
        if (modified == null || this.lastModified == null) {
            this.lastModified = new Date(System.currentTimeMillis());
        } else if (this.lastModified.compareTo(modified) < 0) {
            this.lastModified = modified;
        }
    }

    protected final Node nextInWalk(Node node) {
        // The following looks dirty but makes other things simpler.
        // Most of the time there are whitespace nodes already in the 
        // tree and there's no need to insert an empty text node.
        Node next = node.getNextSibling();
        if (next != null) {
            if (next.getNodeType() != Node.ELEMENT_NODE) {
                return next;
            } else {
                Node realNext = this.template.createTextNode("");
                return next.getParentNode().insertBefore(realNext, next);
            }
        }
        next = this.template.createTextNode("");
        return node.getParentNode().appendChild(next);
    }

    protected Node dispatch(Element elt)
        throws TemplateException, DataSourceException, IOException {
        String ns = elt.getNamespaceURI();
        if (XHTML_NS.equals(ns)) {
            // this is a no-op for XHTML--handling the most common case first
            return elt;
        } else if (TEMPLATE_NS.equals(ns)) {
            String name = elt.getLocalName();
            if ("title".equals(name)) {
                return this.handleTitle(elt);
            } else if ("body".equals(name)) {
                return this.handleBody(elt);
            } else if ("url".equals(name)) {
                return this.handleUrl(elt);
            } else if ("a".equals(name)) {
                return this.handleA(elt);
            } else if ("doc".equals(name)) {
                return this.handleDoc(elt);
            } else if ("foreach".equals(name)) {
                return this.handleForeach(elt);
            } else if ("image".equals(name)) {
                return this.handleImage(elt);
            } else if ("img".equals(name)) {
                return this.handleImg(elt);
            } else if ("date".equals(name)) {
                return this.handleDate(elt);
            } else if ("ifDate".equals(name)) {
                return this.handleIfDate(elt);
            } else if ("switch".equals(name)) {
                return this.handleSwitch(elt);
            } else if ("item".equals(name)) {
                return elt;
            } else if ("found".equals(name)) {
                return elt;
            } else if ("notFound".equals(name)) {
                return elt;
            } else if ("case".equals(name)) {
                return elt;
            } else {
                throw new TemplateException(
                    "found unrecognized element " + elt.getTagName());
            }
        }
        return elt;
    }

    protected void processTemplate()
        throws TemplateException, DataSourceException, IOException {
        this.walkTree(this.template.getDocumentElement());
        /* Ignore the computed lastModified for now */
        this.lastModified = new Date(System.currentTimeMillis());
        this.serialize();
    }

    protected final void walkTree(Node node)
        throws TemplateException, DataSourceException, IOException {
        Node current = node;
        Node next;
        for (;;) {
            if (current.getNodeType() == Node.ELEMENT_NODE) {
                current = this.dispatch((Element)current);
                if ((next = current.getFirstChild()) != null) {
                    current = next;
                    continue;
                }
            }
            for (;;) {
                if ((next = current.getNextSibling()) != null) {
                    current = next;
                    break;
                }
                current = current.getParentNode();
                if (current == node)
                    return;
            }
        }
    }

    protected final boolean needFullContent(Node node)
        throws TemplateException, DataSourceException, IOException {
        Node current = node;
        for (;;) {
            Node next;
            if (current.getNodeType() == Node.ELEMENT_NODE) {
                if (TEMPLATE_NS.equals(current.getNamespaceURI())) {
                    String localName = current.getLocalName();
                    // don't descend into a subtree, if there's a new 
                    // document context there
                    if ("doc".equals(localName)
                        || "foreach".equals(localName)
                        || "image".equals(localName)) {
                        for (;;) {
                            if ((next = current.getNextSibling()) != null) {
                                current = next;
                                break;
                            }
                            current = current.getParentNode();
                            if (current == node)
                                return false;
                        }
                        continue;
                    }
                    if ("switch".equals(localName)) {
                        current = this.handleSwitch((Element)current);
                    } else if (
                        !("title".equals(localName)
                            || "url".equals(localName)
                            || "img".equals(localName)
                            || "date".equals(localName)
                            || "ifDate".equals(localName)
                            || "case".equals(localName)
                            || "a".equals(localName))) {
                        //								need full content if we see a template element which 
                        // isn't in this test
                        return true;
                    }
                }
                if ((next = current.getFirstChild()) != null) {
                    current = next;
                    continue;
                }
            }
            for (;;) {
                if ((next = current.getNextSibling()) != null) {
                    current = next;
                    break;
                }
                current = current.getParentNode();
                if (current == node)
                    return false;
            }
        }
    }

    protected void passThrough()
        throws TemplateException, DataSourceException, IOException {
        Item item = this.dataSource.select(this.context, this.dependencies);
        
        // TODO JK 
        this.response.setDependencies(this.dependencies);
        if (item == null) {
            throw new ResourceNotFoundException();
        } else {
            this.maybeSetLastModified(item.getDateModified());
            this.response.setLastModified(this.lastModified);
            //this.response.setDependencies(this.dependencies);
            if (item instanceof ByteBufferItem) {
                ByteBufferItem byteBuf = (ByteBufferItem)item;
                int length = byteBuf.getContentLength();
                if (length > -1) {
                    this.response.setContentLength(length);
                }
                String contentType = byteBuf.getContentType();
                if (contentType == null) {
                    this.response.setContentType("application/octet-stream");
                } else {
                    this.response.setContentType(contentType);
                }
                byteBuf.writeToOutputStream(response.getOutputStream());
            } else if (item instanceof DocumentItem) {
                this.template = ((DocumentItem)item).getDocument();
                this.serialize();
            } else {
                throw new RuntimeException("No content for passthrough");
            }
        }
    }

    protected void serialize() throws IOException {
        this.response.setLastModified(this.lastModified);
        if(this.dependencies != null) {
            this.response.setDependencies(this.dependencies);
        }
        if (XHTML_NS
            .equals(this.template.getDocumentElement().getNamespaceURI())) {
            this.serializeHTML();
        } else {
            this.serializeXML();
        }
    }

    protected void serializeHTML() throws IOException {
        this.response.setContentType("text/html; charset=utf-8");
        DomParser parser = new DomParser(this.template);
        parser.setContentHandler(
            new HtmlSerializer(this.response.getOutputStream()));
        try {
            parser.setFeature(
                "http://xml.org/sax/features/namespace-prefixes",
                false);
            parser.parse("");
        } catch (SAXException e) {
            if (e.getException() instanceof IOException) {
                throw (IOException)new IOException().initCause(e);
            } else {
                // This shouldn't happen unless there is a bug that can be 
                // likened to a NullPointerException. We're not parsing XML 
                // but traversing a tree that is known to be a tree.
                throw new RuntimeException(e);
            }
        }
    }

    protected void serializeXML() throws IOException {
        if (XHTML_NS
            .equals(this.template.getDocumentElement().getNamespaceURI())) {
            this.response.setContentType("application/xhtml+xml");
        } else {
            this.response.setContentType("application/xml");
        }
        DomParser parser = new DomParser(this.template);
        TextConsumer writer = new TextConsumer(this.response.getOutputStream());
        writer.setXhtml(false);
        writer.setPrettyPrinting(false);
        NSFilter nsFix = new NSFilter(writer);
        TemplateOutputCleaner cleaner = new TemplateOutputCleaner(nsFix);
        parser.setContentHandler(cleaner);
        try {
            parser.setFeature(
                "http://xml.org/sax/features/namespace-prefixes",
                false);
            parser.parse("");
        } catch (SAXException e) {
            if (e.getException() instanceof IOException) {
                throw (IOException)new IOException().initCause(e);
            } else {
                // This shouldn't happen unless there is a bug that can be 
                // likened to a NullPointerException. We're not parsing XML 
                // but traversing a tree that is known to be a tree.
                throw new RuntimeException(e);
            }
        }
    }

    protected Node getCurrentBody()
        throws TemplateException, DataSourceException, IOException {
        Item currItem = this.getCurrentItem();
        if (currItem instanceof DocumentItem) {
            Node body =
                DOMUtils.findBody(((DocumentItem)currItem).getDocument());
            if (body == null) {
                throw new TemplateException("The current document has no <body>");
            }
            return body;
        } else {
            throw new TemplateException("The current document has no document tree");
        }
    }

    protected Node handleTitle(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Node text;
        String title = this.getCurrentItem().getTitle();
        if (title == null || "".equals(title)) {
            // using element content as the default
            title = DOMUtils.textContent(elt);
        }
        text = this.template.createTextNode(title);
        elt.getParentNode().replaceChild(text, elt);
        return text;
    }

    protected Node handleUrl(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Node text;
        URL url = this.getCurrentItem().getURL();
        if (url == null) {
            return elt;
        } else {
            text = this.template.createTextNode(url.toString());
            elt.getParentNode().replaceChild(text, elt);
            return text;
        }
    }

    protected Node handleBody(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Node body = this.getCurrentBody();
        Node parent = elt.getParentNode();
        DOMUtils.importChildrenBefore(body, parent, elt);
        Node next = this.nextInWalk(elt);
        parent.removeChild(elt);
        return next;
    }

    protected Node handleFirstP(Element elt)
        throws TemplateException, DataSourceException, IOException {
        return this.firstNamed(elt, "p");
    }

    protected Node handleH1(Element elt)
        throws TemplateException, DataSourceException, IOException {
        return this.firstNamed(elt, "h1");
    }

    protected Node firstNamed(Element elt, String name)
        throws TemplateException, DataSourceException, IOException {
        Node body = this.getCurrentBody();
        Node namedNode = DOMUtils.findElement(body, XHTML_NS, name);
        if (namedNode == null) {
            return elt;
        } else {
            Node parent = elt.getParentNode();
            DOMUtils.importChildrenBefore(namedNode, parent, elt);
            Node next = this.nextInWalk(elt);
            parent.removeChild(elt);
            return next;
        }
    }

    protected Node handleA(Element elt)
        throws TemplateException, DataSourceException, IOException {
        URL url = this.getCurrentItem().getURL();
        if (url == null) {
            return elt;
        } else {
            Element anchor = this.template.createElementNS(XHTML_NS, "a");
            anchor.setAttribute("href", url.toString());
            String title = this.getCurrentItem().getTitle();
            if ((title != null) && !("".equals(title))) {
                anchor.setAttribute("title", title);
            }
            while (elt.getFirstChild() != null) {
                anchor.appendChild(elt.getFirstChild());
            }
            elt.getParentNode().replaceChild(anchor, elt);
            return anchor;
        }
    }

    protected Node handleImg(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Item item = this.getCurrentItem();
        URL url = item.getURL();
        if ((item instanceof ImageItem) && (url != null)) {
            ImageItem iw = (ImageItem)item;
            int height = iw.getHeight();
            int width = iw.getWidth();
            String alt = iw.getTextualAlternative();
            if (alt == null) {
                alt = "";
            }
            Element img = this.template.createElementNS(XHTML_NS, "img");
            img.setAttribute("src", url.toString());
            if (height > 0 && width > 0) {
                img.setAttribute("height", Integer.toString(height));
                img.setAttribute("width", Integer.toString(width));
            }
            img.setAttribute("alt", alt);
            elt.getParentNode().replaceChild(img, elt);
            return img;
        } else {
            throw new TemplateException("Can't use the current item as an image.");
        }
    }

    protected Node handleDoc(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Node q = null;
        Node found = null;
        Node notFound = null;
        for (Node n = elt.getFirstChild(); n != null; n = n.getNextSibling()) {
            switch (n.getNodeType()) {
                case Node.ELEMENT_NODE :
                    if ("found".equals(n.getLocalName())
                        && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        if (found == null) {
                            found = n;
                        } else {
                            throw new TemplateException("<found> occurred twice in <doc>");
                        }
                    } else if (
                        "notFound".equals(n.getLocalName())
                            && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        if (notFound == null) {
                            notFound = n;
                        } else {
                            throw new TemplateException("<notFound> occurred twice in <doc>");
                        }
                    } else {
                        if (q == null) {
                            q = n;
                        } else {
                            throw new TemplateException("two potential queries in <doc>");
                        }
                    }
                    break;
                case Node.CDATA_SECTION_NODE :
                case Node.TEXT_NODE :
                    if (!DOMUtils.isWhiteSpace(n)) {
                        throw new TemplateException("text content found in <doc>");
                    }
                    break;
                case Node.PROCESSING_INSTRUCTION_NODE :
                    throw new TemplateException("processing instruction found in <doc>");
            }
        }
        if (found == null) {
            throw new TemplateException("<found> missing in <doc>");
        }
        boolean fullContent;
        if (this.avoidFullContent) {
            fullContent = this.needFullContent(found);
        } else {
            fullContent = true;
        }
        Item result;
        if (q == null) {
            result =
                this.getDataSource().select(this.context, this.dependencies);
        } else {
            List resultList =
                this.getDataSource().select(
                    (Element)q,
                    this.context,
                    fullContent,
                    this.dependencies);
            if (resultList.isEmpty()) {
                result = null;
            } else {
                result = (Item)resultList.get(0);
            }
            elt.removeChild(q);
        }
        if (result == null) {
            if (notFound == null) {
                throw new ResourceNotFoundException("empty result");
            } else {
                elt.removeChild(found);
                return notFound;
            }
        } else {
            this.maybeSetLastModified(result.getDateModified());
            this.pushCurrentItem(result);
            this.walkTree(found);
            this.popCurrentItem();
            if (notFound != null) {
                elt.removeChild(notFound);
            }
            return this.nextInWalk(elt);
        }
    }

    protected Node handleImage(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Node found = null;
        Node notFound = null;
        for (Node n = elt.getFirstChild(); n != null; n = n.getNextSibling()) {
            switch (n.getNodeType()) {
                case Node.ELEMENT_NODE :
                    if ("found".equals(n.getLocalName())
                        && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        if (found == null) {
                            found = n;
                        } else {
                            throw new TemplateException("<found> occurred twice in <image>");
                        }
                    } else if (
                        "notFound".equals(n.getLocalName())
                            && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        if (notFound == null) {
                            notFound = n;
                        } else {
                            throw new TemplateException("<notFound> occurred twice in <image>");
                        }
                    } else {
                        throw new TemplateException("only <found> and <notFound> allowed in image");
                    }
                    break;
                case Node.CDATA_SECTION_NODE :
                case Node.TEXT_NODE :
                    if (!DOMUtils.isWhiteSpace(n)) {
                        throw new TemplateException("text content found in <doc>");
                    }
                    break;
                case Node.PROCESSING_INSTRUCTION_NODE :
                    throw new TemplateException("processing instruction found in <doc>");
            }
        }
        if (found == null) {
            throw new TemplateException("<found> missing in <image>");
        }
        Item currentItem = this.getCurrentItem();
        ImageItem result;
        if ((currentItem instanceof IllustratedItem)
            && ((result = ((IllustratedItem)currentItem).getImage()) != null)) {
            this.maybeSetLastModified(result.getDateModified());
            this.pushCurrentItem(result);
            this.walkTree(found);
            this.popCurrentItem();
            if (notFound != null) {
                elt.removeChild(notFound);
            }
            return this.nextInWalk(elt);
        } else {
            elt.removeChild(found);
            if (notFound == null) {
                return this.nextInWalk(elt);
            } else {
                return notFound;
            }
        }
    }

    protected Node handleForeach(Element elt)
        throws TemplateException, DataSourceException, IOException {
        Node q = null;
        Node found = null;
        Node notFound = null;
        for (Node n = elt.getFirstChild(); n != null; n = n.getNextSibling()) {
            switch (n.getNodeType()) {
                case Node.ELEMENT_NODE :
                    if ("found".equals(n.getLocalName())
                        && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        if (found == null) {
                            found = n;
                        } else {
                            throw new TemplateException("<found> occurred twice in <foreach>");
                        }
                    } else if (
                        "notFound".equals(n.getLocalName())
                            && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        if (notFound == null) {
                            notFound = n;
                        } else {
                            throw new TemplateException("<notFound> occurred twice in <foreach>");
                        }
                    } else {
                        if (q == null) {
                            q = n;
                        } else {
                            throw new TemplateException("two potential queries in <foreach>");
                        }
                    }
                    break;
                case Node.CDATA_SECTION_NODE :
                case Node.TEXT_NODE :
                    if (!DOMUtils.isWhiteSpace(n)) {
                        throw new TemplateException("text content found in <foreach>");
                    }
                    break;
                case Node.PROCESSING_INSTRUCTION_NODE :
                    throw new TemplateException("processing instruction found in <foreach>");
            }
        }
        if (found == null) {
            throw new TemplateException("<found> missing in <foreach>");
        }
        if (q == null) {
            throw new TemplateException("<q> missing in <foreach>");
        }
        boolean fullContent;
        if (this.avoidFullContent) {
            fullContent = this.needFullContent(found);
        } else {
            fullContent = true;
        }
        List resultList =
            this.getDataSource().select(
                (Element)q,
                this.context,
                fullContent,
                this.dependencies);
        elt.removeChild(q);
        if (resultList.isEmpty()) {
            if (notFound == null) {
                throw new ResourceNotFoundException("empty result");
            } else {
                elt.removeChild(found);
                return notFound;
            }
        } else {
            Element doc = DOMUtils.findElement(found, TEMPLATE_NS, "item");
            if (doc == null) {
                throw new TemplateException("no <item> in <found> in <foreach>");
            }
            Node docParent = doc.getParentNode();
            Iterator iter = resultList.iterator();
            while (iter.hasNext()) {
                Item result = (Item)iter.next();
                this.maybeSetLastModified(result.getDateModified());
                Node temp = doc.cloneNode(true);
                docParent.insertBefore(temp, doc);
                this.pushCurrentItem(result);
                this.walkTree(temp);
                this.popCurrentItem();
            }
            docParent.removeChild(doc);
            if (notFound != null) {
                elt.removeChild(notFound);
            }
            return this.nextInWalk(elt);
        }
    }

    protected Node handleDate(Element elt)
        throws TemplateException, DataSourceException, IOException {
        String format = elt.getAttribute("format");
        if ("".equals(format)) {
            throw new TemplateException("Date format missing");
        }
        String source = elt.getAttribute("source");
        if ("".equals(source)) {
            source = "modified";
        }
        Date date;
        if ("modified".equals(source)) {
            date = this.getCurrentItem().getDateModified();
        } else {
            int index;
            try {
                index = Integer.parseInt(source);
            } catch (NumberFormatException nfe) {
                throw new TemplateException("Bad date type index", nfe);
            }
            DatedItem di;
            if (this.getCurrentItem() instanceof DatedItem) {
                di = (DatedItem)this.getCurrentItem();
                date = di.getDate(index);
            } else {
                date = null;
            }
        }
        Node parent = elt.getParentNode();
        if (date == null) {
            return elt;
        } else {
            Locale loc = DOMUtils.languageAsLocale(elt);
            SimpleDateFormat formatter;
            if (loc == null) {
                formatter = new SimpleDateFormat(format);
            } else {
                formatter = new SimpleDateFormat(format, loc);
            }
            String formattedDate = formatter.format(date);
            Node text = this.template.createTextNode(formattedDate);
            parent.replaceChild(text, elt);
            return text;
        }
    }

    protected Node handleIfDate(Element elt)
        throws TemplateException, DataSourceException, IOException {
        String source = elt.getAttribute("source");
        Date date;
        if ("modified".equals(source)) {
            date = this.getCurrentItem().getDateModified();
        } else {
            int index;
            try {
                index = Integer.parseInt(source);
            } catch (NumberFormatException nfe) {
                throw new TemplateException("Bad date type index", nfe);
            }
            DatedItem di;
            if (this.getCurrentItem() instanceof DatedItem) {
                di = (DatedItem)this.getCurrentItem();
                date = di.getDate(index);
            } else {
                date = null;
            }
        }
        if (date == null) {
            Node nextInWalk = this.nextInWalk(elt);
            elt.getParentNode().removeChild(elt);
            return nextInWalk;
        } else {
            return elt;
        }
    }

    protected Node handleSwitch(Element elt)
        throws TemplateException, DataSourceException, IOException {
        String path = this.context.getPath();
        Node match = null;
        loop : for (
            Node n = elt.getFirstChild(); n != null; n = n.getNextSibling()) {
            switch (n.getNodeType()) {
                case Node.ELEMENT_NODE :
                    if ("case".equals(n.getLocalName())
                        && TEMPLATE_NS.equals(n.getNamespaceURI())) {
                        String pat = elt.getAttribute("pat");
                        if ("".equals(pat)) {
                            // default
                            match = n;
                            break loop;
                        }
                        try {
                            if (Pattern.matches(pat, path)) {
                                match = n;
                                break loop;
                            }
                        } catch (PatternSyntaxException pse) {
                            throw new TemplateException(
                                "faulty reg exp in <case>",
                                pse);
                        }
                    } else {
                        throw new TemplateException(
                            n.getNodeName() + " found in <switch>");
                    }
                    break;
                case Node.CDATA_SECTION_NODE :
                case Node.TEXT_NODE :
                    if (!DOMUtils.isWhiteSpace(n)) {
                        throw new TemplateException("text content found in <switch>");
                    }
                    break;
                case Node.PROCESSING_INSTRUCTION_NODE :
                    throw new TemplateException("processing instruction found in <switch>");
            }
        }
        if (match == null) {
            Node nextInWalk = this.nextInWalk(elt);
            elt.getParentNode().removeChild(elt);
            return nextInWalk;
        } else {
            Node parent = elt.getParentNode();
            parent.insertBefore(match, elt);
            parent.removeChild(elt);
            return match;
        }
    }
}